from sage.all import *

import base
import utils
import pert

SAVE_CAL_PATH = '/home/wangyi/Dropbox/work/tools/grpy/save/'



class FRW_full():

    def init_vars(self):
        self.var_str = 'Mp k '
        for x in self.coord:
            self.var_str += str(x) + ' '
        self.bg_fun_list = ['a']

        # perturbation vars
        if self.pert_order == 0:
            self.epsilon = 0
            self.pert_fun_list = []
        else:
            self.epsilon = var('epsilon')
            self.var_str += 'epsilon ' 
            self.pert_fun_list = ['psi', 'alpha', 'beta', 'delta_phi']

        if self.pert_order == 2:
            self.psi_k = function('psi_k', self.coord[0])
            self.alpha_k = function('alpha_k', self.coord[0])
            self.beta_k = function('beta_k', self.coord[0])
            self.delta_phi_k = function('delta_phi_k', self.coord[0])

        # define gravity part
        if self.gauge == 'flat':
            print("Chosen flat gauge.")
            self.psi = 0
        else:
            print("Spatial curvature has perturbation psi.")
            self.psi = function('psi', *self.coord)

        self.alpha = function('alpha', *self.coord)
        self.beta = function('beta', *self.coord)

        N0 = 1 + self.epsilon * self.alpha
        N1 = self.epsilon * diff(self.beta,self.coord[1])
        N2 = self.epsilon * diff(self.beta,self.coord[2])
        N3 = self.epsilon * diff(self.beta,self.coord[3])
        AA = self.a**2 * exp(2*self.epsilon*self.psi)
        g00 = -N0**2 + (N1**2 + N2**2 + N3**2)/AA

        self.metric = [[ g00 ,  N1  ,  N2  ,  N3 ],
                       [  N1 ,  AA  ,  0   ,  0  ],
                       [  N2 ,  0   ,  AA  ,  0  ],
                       [  N3 ,  0   ,  0   ,  AA ]]


        if self.matter!=None and self.matter[0]=='scalar':
            # define scalar part
            self.phi0 = function(self.matter[1], self.coord[0])
            self.bg_fun_list.append(self.matter[1])
            if self.gauge == 'uniform_inflaton':
                print('Chosen uniform inflaton gauge.')
                self.delta_phi = 0
            else:
                print('Scalar field has perturbation delta_phi.')
                self.delta_phi = function(self.matter[2], *self.coord)

            self.phi = self.phi0 + self.epsilon * self.delta_phi


    def __init__(self,
                 a=function('a', var('t')),
                 coord=var('t x y z'), 
                 matter=('scalar','phi0','delta_phi'),
                 pert_order = 0,
                 gauge = 'flat',
                 calculate = 'automatic', 
                 prefix_str=SAVE_CAL_PATH+'FRW_'):

        var('epsilon')
        self.coord = coord
        self.simp = lambda x:pert.simp(x, epsilon, pert_order)
        self.pert_order = pert_order
        self.matter = matter
        self.a = a
        self.gauge = gauge

        self.init_vars()

        # calculate or load gravity part
        id_tuple = (self.metric, coord, pert_order)
        load_tuple = utils.load_cal(id_tuple, prefix_str)
        if load_tuple == None or calculate == True:
            curv = base.curvature(self.metric, coord, simp=self.simp)
            self.inverse_metric, self.sqrtg, self.affine,\
                self.riemann, self.ricci, self.R = curv.curvatures
            self.sqrtg_simp = pert.apply_subs(self.sqrtg, self.pert_fun_list, 
                                     coord, pert_order)
            self.sqrtg_simp = self.simp(self.sqrtg_simp)
            self.R_simp = pert.apply_subs(self.R, self.pert_fun_list, 
                                     coord, pert_order)
            self.R_simp = self.simp(self.R_simp)
            utils.save_cal(id_tuple, prefix_str, 
                             (self.inverse_metric, self.sqrtg, self.affine,
                              self.riemann, self.ricci, 
                              self.R, self.sqrtg_simp, self.R_simp))

        else:
            self.inverse_metric, self.sqrtg, self.affine,\
                self.riemann, self.ricci, self.R,\
                self.sqrtg_simp, self.R_simp = load_tuple
            print("Gravity sector loaded from previous calculation.")

        # calculate or load scalar part
        if matter!=None and matter[0]=='scalar':
            id_tuple = (self.gauge, self.metric, coord, pert_order, matter)
            load_tuple = utils.load_cal(id_tuple, prefix_str)
            if load_tuple == None or calculate == True:
                scalar = base.scalar(self.inverse_metric, 
                                     self.affine, self.phi, simp=self.simp)
                self.X = scalar.X
                self.scalar_box = scalar.scalar_box
                self.phi_simp = pert.apply_subs(self.phi, self.pert_fun_list, 
                                     coord, pert_order)
                self.phi_simp = self.simp(self.phi_simp)
                self.X_simp = pert.apply_subs(self.X, self.pert_fun_list, 
                                     coord, pert_order)
                self.X_simp = self.simp(self.X_simp)
                self.scalar_box_simp = pert.apply_subs(self.scalar_box,
                                                       self.pert_fun_list, 
                                                       coord, pert_order)
                self.scalar_box_simp = self.simp(self.scalar_box_simp)
                utils.save_cal(id_tuple, prefix_str,
                               (self.X, self.scalar_box, self.phi_simp,
                                self.X_simp, self.scalar_box_simp))
            else:
                self.X, self.scalar_box, self.phi_simp,\
                    self.X_simp, self.scalar_box_simp = load_tuple


    def cal_Lgrav(self):
        var('Mp')
        lag = self.sqrtg_simp * (Mp**2 * self.R_simp/Integer(2))
        self.Lgrav = self.simp(lag)


    def cal_Lscalar(self):
        V0 = function('V0', self.coord[0])
        V1 = function('V1', self.coord[0])
        V2 = function('V2', self.coord[0])
        V3 = function('V3', self.coord[0])
        V4 = function('V4', self.coord[0])
        self.bg_fun_list.append('V0')
        self.bg_fun_list.append('V1')
        self.bg_fun_list.append('V2')
        self.bg_fun_list.append('V3')
        self.bg_fun_list.append('V4')

        delta_phi = pert.apply_subs(self.delta_phi, self.pert_fun_list, 
                                   self.coord, self.pert_order)

        self.V = V0 + V1 * self.epsilon * delta_phi\
            + V2 * (self.epsilon * delta_phi)**2 /Integer(2) \
            +V3*(self.epsilon*delta_phi)**3/Integer(6)\
            + V4*(self.epsilon*delta_phi)**4/Integer(24)

        lag = self.sqrtg_simp * (self.X_simp - self.V)
        self.Lscalar = self.simp(lag)


    def cal_L(self):
        self.cal_Lgrav()
        self.cal_Lscalar()
        self.L = self.simp(self.Lgrav + self.Lscalar)
        print("Calcalated Lagrangian")
        # print(maxima(self.L))
        # print()
        self.L0 = self.L.coefficient(epsilon,0)
        self.L1 = self.L.coefficient(epsilon,1)
        self.L2 = self.L.coefficient(epsilon,2)


    def cal_int_part(self):
        if self.pert_order != 2:
            print "Pert order != 2 not supported yet."
            return

        try:
            self.L1 + self.L2
        except AttributeError:
            self.cal_L()

        klist=[]
        for elem in self.pert_fun_list:
            klist.append(function(elem+'_k', self.coord[0]))

        ip1 = pert.int_part_se(self.L1, klist, self.coord[0])
        ip1 = pert.int_part_1(ip1, klist, self.coord[0])
        self.L1_int_part = ip1.expand()

        ip2 = pert.int_part_se(self.L2, klist, self.coord[0])
        ip2 = pert.int_part_2(ip2, klist, self.coord[0])
        ip2 = pert.int_part_1(ip2, 
                              [self.alpha_k, self.beta_k], # only constraints
                              self.coord[0])
        self.L2_int_part = ip2.expand()        

        print("Performed integration by parts")
        # print(maxima(self.L2_int_part))
        # print()


    def solve_constraint_eqs(self):
        if self.pert_order != 2:
            print "Pert order != 2 not supported yet."
            return

        try:
            self.L2_int_part
        except AttributeError:
            self.cal_int_part()

        t = self.coord[0]
        var('tmp_a tmp_b')

        # test if alpha and beta has no time derivatives
        if (self.L2_int_part.coeff(diff(self.alpha_k, t))!=0 or
            self.L2_int_part.coeff(diff(self.alpha_k, t,t))!=0):
            print("Cannot solve alpha: still has derivatives.")
            return

        if (self.L2_int_part.coeff(diff(self.beta_k, t))!=0 or
            self.L2_int_part.coeff(diff(self.beta_k, t,t))!=0):
            print("Cannot solve beta: still has derivatives.")
            return

        l2tmp = self.L2_int_part.subs_expr(
            self.alpha_k==tmp_a, self.beta_k==tmp_b)
        l2a = diff(l2tmp, tmp_a)
        self.eq_l2a = l2a
        l2b = diff(l2tmp, tmp_b)
        self.eq_l2b = l2a
        res = solve([l2a==0, l2b==0], tmp_a, tmp_b)
            
        if len(res)!=1:
            print("No solution or non-unique solution of constraints.")
            print(res)
            return

        self.alpha_solution = res[0][0].subs_expr(
            tmp_a==self.alpha_k, tmp_b==self.beta_k)
        self.beta_solution = res[0][1].subs_expr(
            tmp_a==self.alpha_k, tmp_b==self.beta_k)

        print("Solved constraint equations")
        # print(maxima(self.alpha_solution))
        # print(maxima(self.beta_solution))
        # print()


    def cal_L2_solved(self):
        if self.pert_order != 2:
            print "Pert order != 2 not supported yet."
            return

        try:
            self.alpha_solution + self.beta_solution
        except AttributeError:
            self.solve_constraint_eqs()

        print 'Applying alpha solution to L2'
        Lsol = self.L2_int_part.subs_expr(self.alpha_solution)
        Lsol = Lsol.expand()
        print 'Applying beta solution to L2'
        Lsol = Lsol.subs_expr(self.beta_solution)
        Lsol = Lsol.expand()


        print 'Applying some more integ by parts to L2'

        klist=[]
        for elem in self.pert_fun_list:
            klist.append(function(elem+'_k', self.coord[0]))


        Lsol = pert.int_part_se(Lsol, klist, self.coord[0])
        Lsol = pert.int_part_2(Lsol, klist, self.coord[0])

        self.L2_solved = self.simp(Lsol)

        self.L2_solved_collect =self.L2_solved\
            .collect(diff(self.psi_k, self.coord[0])**2)\
            .collect(self.psi_k**2)\
            .collect(diff(self.delta_phi_k, self.coord[0])**2)\
            .collect(self.delta_phi_k**2)

        print("Inserted solution of constraints into L2")
        # print(maxima(self.L2_solved))
        # print()


    def apply_bgeom(self, expr):
        var('Mp')
        x = self.coord
        V0 = function('V0', x[0])
        V1 = function('V1', x[0])
        self.H = function('H',x[0])
        self.bg_fun_list.append('H')
        H = self.H
        a = self.a
        phid = diff(self.phi0,x[0])
        phidd = diff(self.phi0,x[0],x[0])
        Hd = diff(H,x[0])
        Hdd = diff(H,x[0],x[0])

        expr = expr.subs_expr(diff(V0,x[0])==V1*phid) 
        expr = expr.expand()


        expr = expr.subs_expr(
            V0 == 3* Mp**2 * H**2 - phid**2 / Integer(2),
            V1 == - phidd - 3 * H * phid,
            diff(a,x[0]) == a * H,
            diff(a,x[0],x[0]) == a * (H**2 + Hd) ) 
        expr = expr.expand()

        expr = expr.subs_expr(phidd == -Mp**2 * Hdd / phid)
        expr = expr.expand()

        expr = expr.subs_expr(phid**2 == -2 * Mp**2 * Hd)
        expr = expr.subs_expr(phid**4 == 4 * Mp**4 * Hd**2)
        expr = expr.expand()

        print("Warning: Background eom of *standard* model applied.")
        return expr


    def prn_def(self):
        print
        print('# Use the following definitions if needed:')
        # print('sagenb.misc.support.automatic_names(True)')
        print('var("'+self.var_str+'")')
        for elem in self.bg_fun_list:
            print(elem+' = function("'+elem+'", '+str(self.coord[0])+')')
        for elem in self.pert_fun_list:
            print(elem+' = function("'+elem+'", '+str(self.coord[0])+', '
                  +str(self.coord[1])+', '
                  +str(self.coord[2])+', '
                  +str(self.coord[3])+')')
            print(elem+'_k = function("'+elem+'_k", '+str(self.coord[0])+')')
        print

    def prn(self, expr):
        print repr(expr._maxima_()).replace("'","").replace("\\\n","")

# some simple variations of FRW_full:

class FRW(FRW_full):
    def __init__(self,
                 a=function('a', var('t')),
                 coord=var('t x y z'),
                 calculate = 'automatic'):
        FRW_full.__init__(a, coord, matter=None, pert_order = 0,
                          calculate=calculate)

class dS(FRW_full):
    def __init__(self,
                 a=exp(var('H')*var('t')),
                 coord=var('t x y z'),
                 calculate = 'automatic'):
        FRW_full.__init__(self, a, coord, matter=None, pert_order = 0,
                          calculate=calculate)

class FRW_flat_gauge(FRW_full):
    def __init__(self,
                 a=function('a', var('t')),
                 coord=var('t x y z'),
                 calculate = 'automatic'):
        FRW_full.__init__(self, a, coord, matter=('scalar','phi0','delta_phi'),
                          pert_order = 2,
                          gauge = 'flat', 
                          calculate=calculate)

class dS_flat_gauge(FRW_full):
    def __init__(self,
                 a=exp(var('H')*var('t')),
                 coord=var('t x y z'),
                 calculate = 'automatic'):
        FRW_full.__init__(self, a, coord, matter=('scalar','phi0','delta_phi'),
                          pert_order = 2,
                          gauge = 'flat', 
                          calculate=calculate)
        self.var_str += 'H '

class FRW_uniform_inflaton_gauge(FRW_full):
    def __init__(self,
                 a=function('a', var('t')),
                 coord=var('t x y z'),
                 calculate = 'automatic'):
        FRW_full.__init__(self, a, coord, matter=('scalar','phi0','delta_phi'),
                          pert_order = 2,
                          gauge = 'uniform_inflaton', 
                          calculate=calculate)


class dS_uniform_inflaton_gauge(FRW_full):
    def __init__(self,
                 a=exp(var('H')*var('t')),
                 coord=var('t x y z'),
                 calculate = 'automatic'):
        FRW_full.__init__(self, a, coord, matter=('scalar','phi0','delta_phi'),
                          pert_order = 2,
                          gauge = 'uniform_inflaton', 
                          calculate=calculate)
        self.var_str += 'H'


