from sage.all import *

import base
import pert
import models
import utils

class emergon(models.FRW_full):

    def __init__(self,
                 a=function('a', var('t')),
                 coord=var('t x y z'),
                 pert_order = 2,
                 gauge = 'flat',
                 calculate = 'automatic',
                 G_has_pert = True,
                 matter = ('scalar','phi0','delta_phi')
                 ):
        models.FRW_full.__init__(self, 
                                 a, coord, 
                                 matter=matter,
                                 pert_order = pert_order,
                                 gauge=gauge, 
                                 calculate=calculate)
        # init G sector
        self.G0 = function('G', coord[0])
        self.bg_fun_list.append('G')
        self.G = self.G0
        if G_has_pert: 
            self.G = self.G + self.epsilon * function('delta_G', *coord)
            self.pert_fun_list.append('delta_G')
            print "pert list:"
            print self.pert_fun_list

        scalarG = base.scalar(self.inverse_metric, 
                              self.affine, self.G, simp=self.simp)
        self.X_G = scalarG.X
        self.box_G = scalarG.scalar_box
        self.X_G_simp = pert.apply_subs(self.X_G, self.pert_fun_list,
                                        coord, pert_order)
        self.box_G_simp = pert.apply_subs(self.box_G, self.pert_fun_list,
                                          coord, pert_order)

        self.var_str += 'f Lambda xi phi_0 H0 '
        self.prn_def()

    def cal_Lscalar(self):
        var('Mp f Lambda xi epsilon')
        if self.matter == None:
            print("Note: no additional scalar field.")
            self.X_simp = 0
            self.phi_simp = 0
        else:
            print("Note: The scalar sector is different from standard.")
        Lphi = self.sqrtg_simp * exp(2*self.G) * ( 
            self.X_simp - xi * self.R_simp * self.phi_simp**2 /Integer(2))
        Lphi = pert.apply_subs(Lphi, self.pert_fun_list,
                             self.coord, self.pert_order)
        self.Lscalar = self.simp(Lphi)
        
    
    def cal_LG(self):
        var('Mp f Lambda xi epsilon')
        LG = self.sqrtg_simp * ( 
            -2* f**2 * exp(2*self.G) * self.X_G_simp
             - 2* f**3 / Lambda**3 * self.X_G_simp * self.box_G_simp
             + 2* f**3 / Lambda**3 * self.X_G_simp**2
             )
        LG = pert.apply_subs(LG, self.pert_fun_list,
                             self.coord, self.pert_order)
        self.LG = self.simp(LG)

    def cal_L(self):
        self.cal_Lgrav()
        self.cal_Lscalar()
        self.cal_LG()
        self.L = self.simp(self.Lgrav + self.Lscalar + self.LG)
        self.L0 = self.L.coefficient(epsilon,0)
        self.L1 = self.L.coefficient(epsilon,1)
        self.L2 = self.L.coefficient(epsilon,2)


    def apply_bgeom(self, expr):
        var('Mp f xi t H0 k phi_0 Lambda')
        w0 = SR.wild(0)

        # phi_0 moves slowly thus approximately a const
        phi0 = function('phi0', t)

        # b.g. eom for scale factor and G
        a_s = exp(f**2 / (6*Mp**2*H0**2*t**2))
        ad_s = diff(a_s, t)
        add_s = diff(ad_s, t)
        G_s = -log(-H0*t)
        Gd_s = diff(G_s, t)
        Gdd_s = diff(Gd_s, t)

        expr_new = expr.subs_expr(
            self.G0==G_s, diff(self.G0,t)==Gd_s, diff(self.G0,t,t)==Gdd_s, 
            self.a ==a_s, diff(self.a,t)==ad_s, diff(self.a,t,t)==add_s,
            phi0==phi_0, diff(phi0,t)==0, diff(phi0,t,t)==0
            )

        expr_new = expr_new.simplify_full().expand()

        expr_new = expr_new.subs_expr(
            exp(f**2 / (2 * H0**2 * Mp**2 * t**2))==1) # early time approx.

        expr_new = expr_new.simplify_full().expand()

        expr_new = expr_new.subs_expr(
            exp(f**2 / (6 * H0**2 * Mp**2 * t**2))==1) # early time approx.

        expr_new = expr_new.simplify_full().expand()

        expr_new = expr_new.subs_expr(
            exp(f**2 / (3 * H0**2 * Mp**2 * t**2))==1) # early time approx.

        expr_new = expr_new.simplify_full().expand()


        # expr_new = expr_new.subs_expr(
        #     var('Lambda')**3 == ((3*H0**2*f) / Integer(2)))
        expr_new = expr_new.subs_expr(
            var('Lambda') == ((3*H0**2*f) / Integer(2))**(1/Integer(3)))

        expr_new = expr_new.simplify_full().expand()
        return expr_new

    def apply_bgeom_IR(self, expr):
        expr_new = self.apply_bgeom(expr)
        expr_new = expr_new.subs(k==0)
        expr_new = expr_new.expand()
        return expr_new



class emergon_kahler(emergon):


    def cal_Lscalar(self):
        var('Mp f Lambda xi epsilon')
        if self.matter == None:
            print("Note: no additional scalar field.")
            self.X_simp = 0
            self.phi_simp = 0
        else:
            print("Note: The scalar sector is different from standard.")
        Lphi = self.sqrtg_simp * exp(2*self.G) * ( 
            self.X_simp)
        Lphi = pert.apply_subs(Lphi, self.pert_fun_list,
                             self.coord, self.pert_order)
        self.Lscalar = self.simp(Lphi)
        
    
    def cal_LG(self):
        var('Mp f Lambda xi epsilon M')
        LG = self.sqrtg_simp * (1 + self.phi_simp**2 / M**2) * (
            -2* f**2 * exp(2*self.G) * self.X_G_simp
             - 2* f**3 / Lambda**3 * self.X_G_simp * self.box_G_simp
             + 2* f**3 / Lambda**3 * self.X_G_simp**2
             )
        LG = pert.apply_subs(LG, self.pert_fun_list,
                             self.coord, self.pert_order)
        self.LG = self.simp(LG)


class emergon_conformal(models.FRW_full):

    def __init__(self,
                 a=function('a', var('t')),
                 coord=var('t x y z'),
                 pert_order = 2,
                 gauge = 'none',
                 calculate = 'automatic',
                 C_has_pert = True,
                 matter = ('scalar','phi0','delta_phi')
                 ):
        models.FRW_full.__init__(self, 
                                 a, coord, 
                                 matter=matter,
                                 pert_order = pert_order,
                                 gauge=gauge, 
                                 calculate=calculate)

        # init C sector
        self.C0 = function('C', coord[0])
        self.bg_fun_list.append('C')
        self.C = self.C0
        if C_has_pert: 
            self.C = self.C + self.epsilon * function('delta_C', *coord)
            self.pert_fun_list.append('delta_C')
            print "pert list:"
            print self.pert_fun_list

        scalarC = base.scalar(self.inverse_metric, 
                              self.affine, self.C, simp=self.simp)

        self.C_simp = pert.apply_subs(self.C, self.pert_fun_list,
                                        coord, pert_order)
        self.X_C = scalarC.X
        self.X_C_simp = pert.apply_subs(self.X_C, self.pert_fun_list,
                                        coord, pert_order)

        self.var_str += 'M1 M2 xi Mp '
        self.prn_def()


    def cal_L(self):
        var('M1 M2 xi Mp epsilon')
        self.cal_Lgrav()
        self.Lscalars = self.sqrtg_simp * (1 + self.phi_simp**2 / M1**2)*(
            self.X_C_simp + xi * self.C_simp**4 / Integer(4)
            + self.C_simp**2 * self.X_simp / M2**2 )
        
        self.L = self.simp(self.Lgrav + self.Lscalars)
        self.L0 = self.L.coefficient(epsilon,0)
        self.L1 = self.L.coefficient(epsilon,1)
        self.L2 = self.L.coefficient(epsilon,2)


    def apply_bgeom(self, expr):
        var('t M1 M2 xi Mp epsilon phi_0')
        w0 = SR.wild(0)

        # phi_0 moves slowly thus approximately a const
        phi0 = function('phi0', t)

        # b.g. eom for scale factor and C
        a_s = exp(-1/(6 * xi * t**2 * Mp**2))
        ad_s = diff(a_s, t)
        add_s = diff(ad_s, t)
        C_s = - sqrt(2/xi) / t
        Cd_s = diff(C_s, t)
        Cdd_s = diff(Cd_s, t)

        expr_new = expr.subs_expr(
            self.C0==C_s, diff(self.C0,t)==Cd_s, diff(self.C0,t,t)==Cdd_s, 
            self.a ==a_s, diff(self.a,t)==ad_s, diff(self.a,t,t)==add_s,
            phi0==phi_0, diff(phi0,t)==0, diff(phi0,t,t)==0
            )

        expr_new = expr_new.simplify_full().expand()

        expr_new = expr_new.subs_expr(
            exp(-1/(6 * xi * t**2 * Mp**2)) == 1) # early time approx.
        expr_new = expr_new.subs_expr(
            exp(-1/(2 * xi * t**2 * Mp**2)) == 1) # early time approx.
        expr_new = expr_new.subs_expr(
            exp(-1/(3 * xi * t**2 * Mp**2)) == 1) # early time approx.

        expr_new = expr_new.simplify_full().expand()
        return expr_new
