from sage.all import *

def simp(x, epsilon, order):
    if not isinstance(x, Expression):
        x = SR(x)
    x_expand = taylor(x,epsilon,0,order)
    x_coeffs = x_expand.coefficients(epsilon)
    res = 0
    for term in x_coeffs:
        res = res + term[0].expand() * epsilon**term[1]
    return res

def int_part_se(expr, var_list, t):
# f diff(alpha, t, t) --> - diff(f, t) diff(alpha, t), for alpha in list
# f diff(alpha, t) diff(alpha, t, t) -> - diff(f,t) diff(alpha,t)**2 /2
    for v in var_list:
        dv = diff(v,t)
        ddv = diff(v,t,t)
        if dv== 0 or ddv== 0: continue
        term = expr.coefficient(ddv,1)

        if term == 0: 
            continue

        if term.coefficient(ddv)!=0:
            print("Warning: model may have ghost, take care!")
            continue

        if term.coeff(dv)!=0 and term.coeff(dv,2)==0 and term.coeff(dv,3)==0:
            # IMPORTANT: assumed higher order derivatives are also zero
            # linear in diff(v,t) in this case. Now f = sub_term
            sub_term = term.coeff(dv)
            # dsub_term = diff(sub_term, t)
            # expr = expr - sub_term * ddv *dv - dsub_term * dv**2 / Integer(2)
            expr = expr - diff(sub_term * dv**2, t) / Integer(2)
            expr = expr.expand()
            print ('Applied "f dg ddg -> - df dg**2 /2" to  g = '+str(v))
        else:
            dterm = diff(term, t)
            expr = expr - diff(term * dv, t)
            print ('Applied "f ddg  -> - df dg " to  g = '+str(v))

    return expr


def int_part_2(expr, var_list, t):
# f alpha diff(alpha,t) --> - diff(f,t) alpha**2 /2 , for alpha in list
    for v in var_list:
        dv = diff(v,t)
        if dv==0: continue 
        term_tmp = expr.coefficient(v,1)
        term = term_tmp.coefficient(dv,1)
        if term ==0: continue # if no alpha diff(alpha,t) structure
        if term.coefficient(v,1)!=0 or term.coefficient(dv,1)!=0: 
            continue # if there's alpha^2 diff(alpha,t), etc
        # dterm = diff(term,t)
        # expr = expr - term * v * dv - dterm * v**2 / Integer(2)
        expr = expr - diff(term * v**2, t) / Integer(2)
        print ('Applied "f g dg --> - df g**2 /2" to  g = '+str(v))
    return expr


def int_part_1(expr, var_list, t):
# f diff(alpha,t) --> -diff(f,t) alpha, for alpha in list
    for v in var_list:
        dv = diff(v,t)
        term = expr.coefficient(dv,1)
        if term ==0: continue
        # dterm = diff(term, t)
        # expr = expr - term * dv - dterm * v
        expr = expr - diff(term * v, t)
        print ('Applied "f dg --> -df g" to  g = '+str(v))
    return expr


def apply_subs(expr, pert_fun_list, coord, pert_order):
    x = coord
    if len(x)!=4: 
        print("Subs: dim other than 4 not yet supported.")
        return expr

    print('Applying rules to expression...')

    if not isinstance(expr, Expression):
        expr = SR(expr)

    # in the second order case, go to momentum space directly
    if pert_order == 2: 
        var('k')
        for var_str1 in pert_fun_list:
            expr = expr.subs_expr(
                function(var_str1, *x)== function(var_str1+'_k', x[0]))
            expr = expr.subs_expr(
                diff(function(var_str1, *x),x[0])==
                diff(function(var_str1+'_k', x[0]),x[0]))
            expr = expr.subs_expr(
                diff(function(var_str1, *x),x[0],x[0])==
                diff(function(var_str1+'_k', x[0]),x[0],x[0]))

    w0 = SR.wild(0)
    # second order and single function replacements
    for var_str1 in pert_fun_list:
        f = function(var_str1, *x)

        # (partial^2 XX partial^2 XX)
        if pert_order == 2:
            h = k**4 * function(var_str1+'_k',x[0])**2
        else:
            h = function('p2_'+var_str1+'_p2_'+var_str1,*x)
        expr = expr.subs_expr(
            diff(f,x[1],x[1])*diff(f,x[2],x[2])*w0==h*w0/Integer(2)
            - diff(f,x[1],x[1])*diff(f,x[3],x[3])*w0
            - diff(f,x[2],x[2])*diff(f,x[3],x[3])*w0
            - diff(f,x[1],x[1])**2 *w0/Integer(2)
            - diff(f,x[2],x[2])**2 *w0/Integer(2)
            - diff(f,x[3],x[3])**2 *w0/Integer(2))

        # partial_i partial_j XX partial_i partial_j XX
        if pert_order == 2:            
            h = k**4 * function(var_str1+'_k',x[0])**2
        else:
            h = function('pp_'+var_str1+'_pp_'+var_str1,*x)
        expr = expr.subs_expr(
            diff(f,x[1],x[2])**2 *w0==h*w0/Integer(2)
            - diff(f,x[1],x[3])**2 *w0
            - diff(f,x[2],x[3])**2 *w0
            - diff(f,x[1],x[1])**2 *w0/Integer(2)
            - diff(f,x[2],x[2])**2 *w0/Integer(2)
            - diff(f,x[3],x[3])**2 *w0/Integer(2))

    # second order and double function replacements
    for var_str1 in pert_fun_list:
        f = function(var_str1, *x)
        for var_str2 in pert_fun_list:
            g = function(var_str2, *x)

            # (partial_i XX)(partial_i YY)
            if pert_order == 2:
                h = k**2 * function(var_str1+'_k',x[0])\
                    * function(var_str2+'_k',x[0])
            else:
                h = function('p_'+var_str1+'_p_'+var_str2,*x)
            expr = expr.subs_expr( diff(f,x[1])*diff(g,x[1])*w0==h*w0 
                                   - diff(f,x[2])*diff(g,x[2])*w0
                                   - diff(f,x[3])*diff(g,x[3])*w0)

            # (partial_t partial_i XX)(partial_i YY)
            if pert_order == 2:
                h = k**2 * diff(function(var_str1+'_k',x[0]),x[0])\
                    * function(var_str2+'_k',x[0])
            else:
                h = function('p_'+var_str1+'_p_'+var_str2,*x)
            expr = expr.subs_expr( diff(f,x[0],x[1])*diff(g,x[1])*w0==h*w0 
                                   - diff(f,x[0],x[2])*diff(g,x[2])*w0
                                   - diff(f,x[0],x[3])*diff(g,x[3])*w0)

            # (partial_i XX)(partial^2 partial_i YY)
            if pert_order == 2:            
                h = -k**4 * function(var_str1+'_k',x[0])\
                    * function(var_str2+'_k',x[0])
            else:
                h = function('p_'+var_str1+'_pp2_'+var_str2,
                             *x)
            expr = expr.subs_expr(
                diff(f,x[1])*diff(g,x[1],x[1],x[1])*w0==h*w0 
                - diff(f,x[1])*diff(g,x[1],x[2],x[2])*w0
                - diff(f,x[1])*diff(g,x[1],x[3],x[3])*w0
                - diff(f,x[2])*diff(g,x[2],x[1],x[1])*w0
                - diff(f,x[2])*diff(g,x[2],x[2],x[2])*w0
                - diff(f,x[2])*diff(g,x[2],x[3],x[3])*w0
                - diff(f,x[3])*diff(g,x[3],x[1],x[1])*w0
                - diff(f,x[3])*diff(g,x[3],x[2],x[2])*w0
                - diff(f,x[3])*diff(g,x[3],x[3],x[3])*w0)

    # first order replacements
    for var_str1 in pert_fun_list:
        f = function(var_str1, *x)
        # (partial^2 XX)
        if pert_order == 2:            
            h = -k**2 * function(var_str1+'_k',x[0])
        else:
            h = function('p2_'+var_str1,*x)
        expr = expr.subs_expr(diff(f,x[1],x[1])*w0==h*w0
                              -diff(f,x[2],x[2])*w0-diff(f,x[3],x[3])*w0)
        # (partial^2 partial_t XX)
        # h is the same as above. don't insert another h!
        expr = expr.subs_expr(diff(f,x[0],x[1],x[1])*w0==
                              diff(h,x[0])*w0
                              -diff(f,x[0],x[2],x[2])*w0
                              -diff(f,x[0],x[3],x[3])*w0)

    return expr





